import functools
import http.server
import os
import pathlib
import shutil
import signal
import socketserver
import subprocess
import sys
import tempfile
import threading
import webbrowser


from jinja2 import Environment, FileSystemLoader
from types import BuiltinFunctionType, FrameType, FunctionType, ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast

from scalene.scalene_statistics import (
    Filename,
    LineNumber,
    StackFrame,
    StackStats,
    ScaleneStatistics,
)
from scalene.scalene_config import scalene_version, scalene_date


def enter_function_meta(
    frame: FrameType,
    should_trace: Callable[[Filename, str], bool],
    stats: ScaleneStatistics,
) -> None:
    """Update tracking info so we can correctly report line number info later."""
    fname = Filename(frame.f_code.co_filename)
    lineno = LineNumber(frame.f_lineno)

    f = frame
    try:
        while "<" in Filename(f.f_code.co_name):
            f = cast(FrameType, f.f_back)
            # Handle case where the function with the name wrapped
            # in triangle brackets is at the bottom of the stack
            if f is None:
                return
    except Exception:
        return
    if not should_trace(Filename(f.f_code.co_filename), f.f_code.co_name):
        return

    fn_name = get_fully_qualified_name(f)
    firstline = f.f_code.co_firstlineno

    stats.function_map[fname][lineno] = fn_name
    stats.firstline_map[fn_name] = LineNumber(firstline)


def compute_frames_to_record(
    should_trace: Callable[[Filename, str], bool],
) -> List[Tuple[FrameType, int, FrameType]]:
    """Collect all stack frames that Scalene actually processes.
    Returns final frame (up to a line in a file we are profiling), the
    thread identifier, and the original frame.
    """
    frames: List[Tuple[FrameType, int]] = [
        (
            cast(
                FrameType,
                sys._current_frames().get(cast(int, t.ident), None),
            ),
            cast(int, t.ident),
        )
        for t in threading.enumerate()
        if t != threading.main_thread()
    ]
    # Put the main thread in the front.

    tid = cast(int, threading.main_thread().ident)
    frames.insert(
        0,
        (
            sys._current_frames().get(tid, cast(FrameType, None)),
            tid,
        ),
    )
    # Process all the frames to remove ones we aren't going to track.
    new_frames: List[Tuple[FrameType, int, FrameType]] = []
    for frame, tident in frames:
        orig_frame = frame
        if not frame:
            continue
        fname = frame.f_code.co_filename
        func = frame.f_code.co_name
        # Record samples only for files we care about.
        if not fname:
            # 'eval/compile' gives no f_code.co_filename.  We have
            # to look back into the outer frame in order to check
            # the co_filename.
            back = cast(FrameType, frame.f_back)
            fname = Filename(back.f_code.co_filename)
            func = back.f_code.co_name
        while not should_trace(Filename(fname), func):
            # Walk the stack backwards until we hit a frame that
            # IS one we should trace (if there is one).  i.e., if
            # it's in the code being profiled, and it is just
            # calling stuff deep in libraries.
            if frame:
                frame = cast(FrameType, frame.f_back)
            else:
                break
            if frame:
                fname = frame.f_code.co_filename
                func = frame.f_code.co_name
        if frame:
            new_frames.append((frame, tident, orig_frame))
    del frames[:]
    return new_frames


def add_stack(
    frame: FrameType,
    should_trace: Callable[[Filename, str], bool],
    stacks: Dict[Tuple[StackFrame, ...], StackStats],
    python_time: float,
    c_time: float,
    cpu_samples: float,
) -> None:
    """Add one to the stack starting from this frame."""
    stk: List[StackFrame] = list()
    f: Optional[FrameType] = frame
    while f:
        if should_trace(Filename(f.f_code.co_filename), f.f_code.co_name):
            stk.insert(
                0,
                StackFrame(
                    filename=str(f.f_code.co_filename),
                    function_name=str(get_fully_qualified_name(f)),
                    line_number=int(f.f_lineno),
                ),
            )
        f = f.f_back
    stack_tuple = tuple(stk)
    if stack_tuple not in stacks:
        stacks[stack_tuple] = StackStats(1, python_time, c_time, cpu_samples)
    else:
        prev_stats = stacks[stack_tuple]
        stacks[stack_tuple] = StackStats(
            prev_stats.count + 1,
            prev_stats.python_time + python_time,
            prev_stats.c_time + c_time,
            prev_stats.cpu_samples + cpu_samples,
        )


def on_stack(
    frame: FrameType, fname: Filename, lineno: LineNumber
) -> Optional[FrameType]:
    """Find a frame matching the given filename and line number, if any.

    Used for checking whether we are still executing the same line
    of code or not in invalidate_lines (for per-line memory
    accounting).
    """
    f = frame
    current_file_and_line = (fname, lineno)
    while f:
        if (f.f_code.co_filename, f.f_lineno) == current_file_and_line:
            return f
        f = cast(FrameType, f.f_back)
    return None


def get_fully_qualified_name(frame: FrameType) -> Filename:
    # Obtain the fully-qualified name.
    if sys.version_info >= (3, 11):
        # Introduced in Python 3.11
        fn_name = Filename(frame.f_code.co_qualname)
        return fn_name
    f = frame
    # Manually search for an enclosing class.
    fn_name = Filename(f.f_code.co_name)
    while f and f.f_back and f.f_back.f_code:
        if "self" in f.f_locals:
            prepend_name = f.f_locals["self"].__class__.__name__
            if "Scalene" not in prepend_name:
                fn_name = Filename(f"{prepend_name}.{fn_name}")
            break
        if "cls" in f.f_locals:
            prepend_name = getattr(f.f_locals["cls"], "__name__", None)
            if not prepend_name or "Scalene" in prepend_name:
                break
            fn_name = Filename(f"{prepend_name}.{fn_name}")
            break
        f = f.f_back
    return fn_name


def flamegraph_format(stacks: Dict[Tuple[StackFrame], StackStats]) -> str:
    """Converts stacks to a string suitable for input to Brendan Gregg's flamegraph.pl script."""
    output = ""
    for stk in stacks:
        for frame in stk:
            output += f"{frame.filename} {frame.function_name}:{frame.line_number};"
        output += " " + str(stacks[stk].count)
        output += "\n"
    return output


def generate_html(profile_fname: Filename, output_fname: Filename) -> None:
    """Apply a template to generate a single HTML payload containing the current profile."""

    def read_file_content(directory: str, subdirectory: str, filename: str) -> str:
        file_path = os.path.join(directory, subdirectory, filename)
        file_content = ""
        try:
            file_content = pathlib.Path(file_path).read_text(encoding="utf-8")
        except UnicodeDecodeError as e:
            # Create a new error with just the custom message
            raise UnicodeDecodeError(
                "utf-8",
                b"",
                0,
                0,
                f"Failed to decode file {file_path}. Ensure the file is UTF-8 encoded.",
            ) from e
        return file_content

    try:
        # Load the profile
        profile_file = pathlib.Path(profile_fname)
        profile = ""
        try:
            profile = profile_file.read_text(encoding="utf-8")
        except UnicodeDecodeError as e:
            # Create a new error with just the custom message
            raise UnicodeDecodeError(
                "utf-8",
                b"",
                0,
                0,
                f"Failed to decode file {profile_file}. Ensure the file is UTF-8 encoded.",
            ) from e

    except FileNotFoundError:
        assert profile_fname == "demo"
        profile = ""
        # return

    # Load the GUI JavaScript file.
    scalene_dir = os.path.dirname(__file__)

    file_contents = {
        "scalene_gui_js_text": read_file_content(
            scalene_dir, "scalene-gui", "scalene-gui-bundle.js"
        ),
        "prism_css_text": read_file_content(scalene_dir, "scalene-gui", "prism.css"),
    }

    # Put the profile and everything else into the template.
    environment = Environment(
        loader=FileSystemLoader(os.path.join(scalene_dir, "scalene-gui"))
    )
    template = environment.get_template("index.html.template")
    rendered_content = template.render(
        profile=profile,
        gui_js=file_contents["scalene_gui_js_text"],
        prism_css=file_contents["prism_css_text"],
        scalene_version=scalene_version,
        scalene_date=scalene_date,
    )

    # Write the rendered content to the specified output file.
    try:
        with open(output_fname, "w", encoding="utf-8") as f:
            f.write(rendered_content)
    except OSError:
        pass


def start_server(port: int, directory: str) -> None:
    try:
        handler = http.server.SimpleHTTPRequestHandler
        with socketserver.TCPServer(("", port), handler) as httpd:
            os.chdir(directory)
            httpd.serve_forever()
    except OSError:
        # print(f"Port {port} is already in use. Please try a different port.")
        pass


def show_browser(file_path: str, port: int, orig_python: str = "python3") -> None:
    temp_dir = tempfile.gettempdir()

    # Copy file to the temporary directory
    shutil.copy(file_path, os.path.join(temp_dir, "index.html"))

    # Open web browser in a new subprocess
    curr_dir = os.getcwd()
    try:
        os.chdir(temp_dir)
        subprocess.Popen(
            [
                orig_python,
                os.path.join(os.path.dirname(__file__), "launchbrowser.py"),
                file_path,
                f"{port}",
            ]
        )
        # Open web browser to local server
        webbrowser.open(f"http://localhost:{port}/")
    except (FileNotFoundError, PermissionError, OSError):
        pass
    except webbrowser.Error:
        pass
    finally:
        os.chdir(curr_dir)


def patch_module_functions_with_signal_blocking(
    module: ModuleType, signal_to_block: signal.Signals
) -> None:
    """Patch all functions in the given module to block the specified signal during execution."""

    def signal_blocking_wrapper(func: Union[BuiltinFunctionType, FunctionType]) -> Any:
        """Wrap a function to block the specified signal during its execution."""

        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            # Block the specified signal temporarily
            original_sigmask = signal.pthread_sigmask(
                signal.SIG_BLOCK, [signal_to_block]
            )
            try:
                return func(*args, **kwargs)
            finally:
                # Restore original signal mask
                signal.pthread_sigmask(signal.SIG_SETMASK, original_sigmask)

        return wrapped

    # Iterate through all attributes of the module
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if isinstance(attr, (BuiltinFunctionType, FunctionType)):
            wrapped_attr = signal_blocking_wrapper(attr)
            setattr(module, attr_name, wrapped_attr)



# file scalene/__main__.py:13-21
# lines [13, 14, 15, 17, 18, 19, 20, 21]
# branches []

import pytest
from unittest.mock import patch
import sys
import io
from scalene import scalene_profiler


# Test function to improve coverage for the main function in scalene.__main__
def test_main_exception_handling():
    # Mock the Scalene main function to raise an exception
    with patch(
        "scalene.scalene_profiler.Scalene.main", side_effect=Exception("Test Exception")
    ):
        # Redirect stderr to capture the output
        with patch("sys.stderr", new=io.StringIO()) as fake_stderr:
            # Mock sys.exit to prevent the test from exiting
            with patch("sys.exit", side_effect=SystemExit) as mock_exit:
                # Call the main function which should now raise an exception
                with pytest.raises(SystemExit):
                    from scalene.__main__ import main

                    main()
                # Check that the exception message was printed to stderr
                assert (
                    "ERROR: Calling Scalene main function failed: Test Exception"
                    in fake_stderr.getvalue()
                )
                # Check that sys.exit was called with the correct exit code
                mock_exit.assert_called_once_with(1)



#!/usr/bin/env python3
import json
import pathlib
import tempfile
import subprocess
import sys


def smoketest(fname, rest):
    outfile = pathlib.Path(
        tempfile.mkdtemp(prefix="scalene") / pathlib.Path("smoketest.json")
    )
    cmd = [
        sys.executable,
        "-m",
        "scalene",
        "--cli",
        "--json",
        "--outfile",
        str(outfile),
        *rest,
        fname,
    ]
    print("COMMAND", " ".join(cmd))
    proc = subprocess.run(cmd, capture_output=True)
    stdout = proc.stdout.decode("utf-8")
    stderr = proc.stderr.decode("utf-8")

    if proc.returncode != 0:
        print("Exited with a non-zero code:", proc.returncode)
        print("STDOUT", stdout)
        print("STDERR", stderr)

        exit(proc.returncode)
    #    print("STDOUT", stdout)
    #    print("\nSTDERR", stderr)
    try:
        with open(outfile, "r") as f:
            outfile_contents = f.read()
        scalene_json = json.loads(outfile_contents)
    except json.JSONDecodeError:
        print("Invalid JSON", stderr)
        print("STDOUT", stdout)
        print("STDERR", stderr)
        exit(1)
    if len(scalene_json) == 0:
        print("No JSON output")
        print("STDOUT", stdout)
        print("STDERR", stderr)
        exit(1)
    files = scalene_json["files"]
    if not len(files) > 0:
        print("No files found in output")
        exit(1)
    for _fname in files:

        if not any(
            (line["n_cpu_percent_c"] > 0 or line["n_cpu_percent_python"] > 0)
            for line in files[_fname]["lines"]
        ):
            print("No non-zero lines in", _fname)
            exit(1)


if __name__ == "__main__":
    smoketest(sys.argv[1], sys.argv[2:])



import copy
import math
import random
import re

from collections import defaultdict
from enum import Enum
from operator import itemgetter
from pathlib import Path
from pydantic import (
    BaseModel,
    Field,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
    StrictBool,
    ValidationError,
    model_validator,
)
from typing import Any, Callable, Dict, List, Optional

from scalene.scalene_leak_analysis import ScaleneLeakAnalysis
from scalene.scalene_statistics import (
    Filename,
    LineNumber,
    ScaleneStatistics,
    StackStats,
)
from scalene.scalene_analysis import ScaleneAnalysis


class GPUDevice(str, Enum):
    nvidia = "GPU"
    neuron = "Neuron"
    no_gpu = ""


class FunctionDetail(BaseModel):
    line: str
    lineno: LineNumber
    memory_samples: List[List[Any]]
    n_avg_mb: NonNegativeFloat
    n_copy_mb_s: NonNegativeFloat
    n_core_utilization: float = Field(..., ge=0, le=1)
    cpu_samples_list: List[float]
    n_cpu_percent_c: float = Field(..., ge=0, le=100)
    n_cpu_percent_python: float = Field(..., ge=0, le=100)
    n_gpu_avg_memory_mb: NonNegativeFloat
    n_gpu_peak_memory_mb: NonNegativeFloat
    n_gpu_percent: float = Field(..., ge=0, le=100)
    n_growth_mb: NonNegativeFloat
    n_peak_mb: NonNegativeFloat
    n_malloc_mb: NonNegativeFloat
    n_mallocs: NonNegativeInt
    n_python_fraction: float = Field(..., ge=0, le=1)
    n_sys_percent: float = Field(..., ge=0, le=100)
    n_usage_fraction: float = Field(..., ge=0, le=1)

    @model_validator(mode="after")
    def check_cpu_percentages(self) -> Any:
        total_cpu_usage = math.floor(
            self.n_cpu_percent_c + self.n_cpu_percent_python + self.n_sys_percent
        )
        if total_cpu_usage > 100:
            raise ValueError(
                f"The sum of n_cpu_percent_c, n_cpu_percent_python, and n_sys_percent must be <= 100 but is {total_cpu_usage}"
            )
        return self

    @model_validator(mode="after")
    def check_gpu_memory(self) -> Any:
        if self.n_gpu_avg_memory_mb > self.n_gpu_peak_memory_mb:
            raise ValueError(
                "n_gpu_avg_memory_mb must be less than or equal to n_gpu_peak_memory_mb"
            )
        return self

    @model_validator(mode="after")
    def check_cpu_memory(self) -> Any:
        if self.n_avg_mb > self.n_peak_mb:
            raise ValueError("n_avg_mb must be less than or equal to n_peak_mb")
        return self


class LineDetail(FunctionDetail):
    start_outermost_loop: PositiveInt
    end_outermost_loop: PositiveInt
    start_region_line: PositiveInt
    end_region_line: PositiveInt


class LeakInfo(BaseModel):
    likelihood: NonNegativeFloat
    velocity_mb_s: NonNegativeFloat


class FileDetail(BaseModel):
    functions: List[FunctionDetail]
    imports: List[str]
    leaks: Dict[str, LeakInfo]
    lines: List[LineDetail]
    percent_cpu_time: NonNegativeFloat


class ScaleneJSONSchema(BaseModel):
    alloc_samples: NonNegativeInt
    args: Optional[List[str]] = None
    elapsed_time_sec: NonNegativeFloat
    start_time_absolute: NonNegativeFloat
    start_time_perf: NonNegativeFloat
    entrypoint_dir: str
    filename: str
    files: Dict[str, FileDetail]
    gpu: StrictBool
    gpu_device: GPUDevice
    growth_rate: float
    max_footprint_fname: Optional[str]
    max_footprint_lineno: Optional[PositiveInt]
    max_footprint_mb: NonNegativeFloat
    max_footprint_python_fraction: NonNegativeFloat
    memory: StrictBool
    program: str
    samples: List[List[NonNegativeFloat]]
    stacks: List[List[Any]]


class ScaleneJSON:
    @staticmethod
    def memory_consumed_str(size_in_mb: float) -> str:
        """Return a string corresponding to amount of memory consumed."""
        gigabytes = size_in_mb // 1024
        terabytes = gigabytes // 1024
        if terabytes > 0:
            return f"{(size_in_mb / 1048576):3.3f} TB"
        elif gigabytes > 0:
            return f"{(size_in_mb / 1024):3.3f} GB"
        else:
            return f"{size_in_mb:3.3f} MB"

    @staticmethod
    def time_consumed_str(time_in_ms: float) -> str:
        hours = time_in_ms // 3600000
        minutes = (time_in_ms % 3600000) // 60000
        seconds = (time_in_ms % 60000) // 1000
        hours_exact = time_in_ms / 3600000
        minutes_exact = (time_in_ms % 3600000) / 60000
        seconds_exact = (time_in_ms % 60000) / 1000
        if hours > 0:
            return f"{hours_exact:.0f}h:{minutes_exact:.0f}m:{seconds_exact:3.3f}s"
        elif minutes > 0:
            return f"{minutes_exact:.0f}m:{seconds_exact:3.3f}s"
        elif seconds > 0:
            return f"{seconds_exact:3.3f}s"
        else:
            return f"{time_in_ms:3.3f}ms"

    # Default threshold for percent of CPU time to report a file.
    cpu_percent_threshold = 1

    # Default threshold for number of mallocs to report a file.
    malloc_threshold = 1  # 100

    # Fraction of the maximum footprint to use as granularity for memory timelines
    # (used for compression). E.g., 10 => 1/10th of the max.
    memory_granularity_fraction = 10

    # Maximum number of sparkline samples.
    max_sparkline_samples = 100

    def __init__(self) -> None:
        # where we write profile info
        self.output_file = ""

        # if we are on a GPU or not
        self.gpu = False
        self.gpu_device = ""

    def compress_samples(self, samples: List[Any], max_footprint: float) -> Any:
        if len(samples) <= self.max_sparkline_samples:
            return samples

        new_samples = sorted(
            random.sample(list(map(tuple, samples)), self.max_sparkline_samples)
        )
        return new_samples

    # Profile output methods
    def output_profile_line(
        self,
        *,
        fname: Filename,
        fname_print: Filename,
        line_no: LineNumber,
        line: str,
        stats: ScaleneStatistics,
        profile_this_code: Callable[[Filename, LineNumber], bool],
        profile_memory: bool = False,
        force_print: bool = False,
    ) -> Dict[str, Any]:
        """Print at most one line of the profile (true == printed one)."""

        if not force_print and not profile_this_code(fname, line_no):
            return {
                "lineno": line_no,
                "line": line,
                "cpu_samples_list": [],
                "n_core_utilization": 0,
                "n_cpu_percent_c": 0,
                "n_cpu_percent_python": 0,
                "n_sys_percent": 0,
                "n_gpu_percent": 0,
                "n_gpu_avg_memory_mb": 0,
                "n_gpu_peak_memory_mb": 0,
                "n_peak_mb": 0,
                "n_growth_mb": 0,
                "n_avg_mb": 0,
                "n_mallocs": 0,
                "n_malloc_mb": 0,
                "n_usage_fraction": 0,
                "n_python_fraction": 0,
                "n_copy_mb_s": 0,
                "memory_samples": [],
            }

        # Prepare output values.
        n_cpu_samples_c = stats.cpu_stats.cpu_samples_c[fname][line_no]
        # Correct for negative CPU sample counts. This can happen
        # because of floating point inaccuracies, since we perform
        # subtraction to compute it.
        n_cpu_samples_c = max(0, n_cpu_samples_c)
        n_cpu_samples_python = stats.cpu_stats.cpu_samples_python[fname][line_no]
        n_gpu_samples = stats.gpu_stats.gpu_samples[fname][line_no]
        n_gpu_mem_samples = stats.gpu_stats.gpu_mem_samples[fname][line_no]

        # Compute percentages of CPU time.
        if stats.cpu_stats.total_cpu_samples:
            n_cpu_percent_c = n_cpu_samples_c * 100 / stats.cpu_stats.total_cpu_samples
            n_cpu_percent_python = (
                n_cpu_samples_python * 100 / stats.cpu_stats.total_cpu_samples
            )
        else:
            n_cpu_percent_c = 0
            n_cpu_percent_python = 0

        if True:
            if stats.gpu_stats.n_gpu_samples[fname][line_no]:
                n_gpu_percent = (
                    n_gpu_samples * 100 / stats.gpu_stats.n_gpu_samples[fname][line_no]
                )  # total_gpu_samples
            else:
                n_gpu_percent = 0

        # Now, memory stats.
        # Total volume of memory allocated.
        n_malloc_mb = stats.memory_stats.memory_malloc_samples[fname][line_no]
        # Number of distinct allocation calls (those from the same line are counted as 1).
        n_mallocs = stats.memory_stats.memory_malloc_count[fname][line_no]
        # Total volume of memory allocated by Python (not native code).
        n_python_malloc_mb = stats.memory_stats.memory_python_samples[fname][line_no]

        n_usage_fraction = (
            0
            if not stats.memory_stats.total_memory_malloc_samples
            else n_malloc_mb / stats.memory_stats.total_memory_malloc_samples
        )
        n_python_fraction = 0 if not n_malloc_mb else n_python_malloc_mb / n_malloc_mb

        # Average memory consumed by this line.
        n_avg_mb = (
            stats.memory_stats.memory_aggregate_footprint[fname][line_no]
            if n_mallocs == 0
            else stats.memory_stats.memory_aggregate_footprint[fname][line_no]
            / n_mallocs
        )

        # Peak memory consumed by this line.
        n_peak_mb = stats.memory_stats.memory_max_footprint[fname][line_no]

        # Force the reporting of average to be no more than peak.
        # In principle, this should never happen, but...
        # assert n_avg_mb <= n_peak_mb
        if n_avg_mb > n_peak_mb:
            n_avg_mb = n_peak_mb

        n_cpu_percent = n_cpu_percent_c + n_cpu_percent_python

        # Adjust CPU time by utilization.
        mean_cpu_util = stats.cpu_stats.cpu_utilization[fname][line_no].mean()
        mean_core_util = stats.cpu_stats.core_utilization[fname][line_no].mean()
        n_sys_percent = n_cpu_percent * (1.0 - mean_cpu_util)
        n_cpu_percent_python *= mean_cpu_util
        n_cpu_percent_c *= mean_cpu_util
        del mean_cpu_util

        n_copy_b = stats.memory_stats.memcpy_samples[fname][line_no]
        if stats.elapsed_time:
            n_copy_mb_s = n_copy_b / (1024 * 1024 * stats.elapsed_time)
        else:
            n_copy_mb_s = 0

        stats.memory_stats.per_line_footprint_samples[fname][line_no] = (
            self.compress_samples(
                stats.memory_stats.per_line_footprint_samples[fname][line_no],
                stats.memory_stats.max_footprint,
            )
        )

        payload = {
            "line": line,
            "lineno": line_no,
            "memory_samples": stats.memory_stats.per_line_footprint_samples[fname][
                line_no
            ],
            "cpu_samples_list": stats.cpu_stats.cpu_samples_list[fname][line_no],
            "n_avg_mb": n_avg_mb,
            "n_copy_mb_s": n_copy_mb_s,
            "n_core_utilization": mean_core_util,
            "n_cpu_percent_c": n_cpu_percent_c,
            "n_cpu_percent_python": n_cpu_percent_python,
            "n_gpu_avg_memory_mb": n_gpu_mem_samples.mean(),
            "n_gpu_peak_memory_mb": n_gpu_mem_samples.peak(),
            "n_gpu_percent": n_gpu_percent,
            "n_growth_mb": n_peak_mb,  # For backwards compatibility
            "n_peak_mb": n_peak_mb,
            "n_malloc_mb": n_malloc_mb,
            "n_mallocs": n_mallocs,
            "n_python_fraction": n_python_fraction,
            "n_sys_percent": n_sys_percent,
            "n_usage_fraction": n_usage_fraction,
        }
        try:
            FunctionDetail(**payload)
        except ValidationError as e:
            print("Warning: JSON failed validation:")
            print(e)
        return payload

    def output_profiles(
        self,
        program: Filename,
        stats: ScaleneStatistics,
        pid: int,
        profile_this_code: Callable[[Filename, LineNumber], bool],
        python_alias_dir: Path,
        program_path: Filename,
        entrypoint_dir: Filename,
        program_args: Optional[List[str]],
        profile_memory: bool = True,
        reduced_profile: bool = False,
    ) -> Dict[str, Any]:
        """Write the profile out."""
        # Get the children's stats, if any.
        if not pid:
            stats.merge_stats(python_alias_dir)
        # If we've collected any samples, dump them.
        if (
            not stats.cpu_stats.total_cpu_samples
            and not stats.memory_stats.total_memory_malloc_samples
            and not stats.memory_stats.total_memory_free_samples
            and not stats.gpu_stats.n_gpu_samples[program]
        ):
            # Nothing to output.
            return {}
        # Collect all instrumented filenames.
        all_instrumented_files: List[Filename] = list(
            set(
                list(stats.cpu_stats.cpu_samples_python.keys())
                + list(stats.cpu_stats.cpu_samples_c.keys())
                + list(stats.memory_stats.memory_free_samples.keys())
                + list(stats.memory_stats.memory_malloc_samples.keys())
                + list(stats.gpu_stats.gpu_samples.keys())
            )
        )
        if not all_instrumented_files:
            # We didn't collect samples in source files.
            return {}
        growth_rate = 0.0
        if profile_memory:
            stats.memory_stats.memory_footprint_samples = self.compress_samples(
                stats.memory_stats.memory_footprint_samples,
                stats.memory_stats.max_footprint,
            )

            # Compute growth rate (slope), between 0 and 1.
            if stats.memory_stats.allocation_velocity[1] > 0:
                growth_rate = (
                    100.0
                    * stats.memory_stats.allocation_velocity[0]
                    / stats.memory_stats.allocation_velocity[1]
                )
        else:
            stats.memory_stats.memory_footprint_samples = []

        # Adjust the program name if it was a Jupyter cell.
        result = re.match(r"_ipython-input-([0-9]+)-.*", program)
        if result:
            program = Filename("[" + result.group(1) + "]")

        # Process the stacks to normalize by total number of CPU samples.
        for stk in stats.stacks:
            stack_stats = stats.stacks[stk]
            stats.stacks[stk] = StackStats(
                stack_stats.count,
                stack_stats.python_time / stats.cpu_stats.total_cpu_samples,
                stack_stats.c_time / stats.cpu_stats.total_cpu_samples,
                stack_stats.cpu_samples / stats.cpu_stats.total_cpu_samples,
            )

        # Convert stacks into a representation suitable for JSON dumping.
        stks = []
        for stk in stats.stacks:
            this_stk: List[str] = []
            this_stk.extend(str(frame) for frame in stk)
            stack_stats = stats.stacks[stk]
            # Convert StackStats to a dictionary
            stack_stats_dict = {
                "count": stack_stats.count,
                "python_time": stack_stats.python_time,
                "c_time": stack_stats.c_time,
                "cpu_samples": stack_stats.cpu_samples,
            }
            stks.append((this_stk, stack_stats_dict))

        output: Dict[str, Any] = {
            "program": program,
            "entrypoint_dir": entrypoint_dir,
            "args": program_args,
            "filename": program_path,
            "alloc_samples": stats.memory_stats.alloc_samples,
            "elapsed_time_sec": stats.elapsed_time,
            "start_time_absolute": stats.start_time_absolute,
            "start_time_perf": stats.start_time_perf,
            "growth_rate": growth_rate,
            "max_footprint_mb": stats.memory_stats.max_footprint,
            "max_footprint_python_fraction": stats.memory_stats.max_footprint_python_fraction,
            "max_footprint_fname": (
                stats.memory_stats.max_footprint_loc[0]
                if stats.memory_stats.max_footprint_loc
                else None
            ),
            "max_footprint_lineno": (
                stats.memory_stats.max_footprint_loc[1]
                if stats.memory_stats.max_footprint_loc
                else None
            ),
            "files": {},
            "gpu": self.gpu,
            "gpu_device": self.gpu_device,
            "memory": profile_memory,
            "samples": stats.memory_stats.memory_footprint_samples,
            "stacks": stks,
        }

        # Build a list of files we will actually report on.
        report_files: List[Filename] = []
        # Sort in descending order of CPU cycles, and then ascending order by filename
        for fname in sorted(
            all_instrumented_files,
            key=lambda f: (-(stats.cpu_stats.cpu_samples[f]), f),
        ):
            fname = Filename(fname)
            try:
                percent_cpu_time = (
                    100
                    * stats.cpu_stats.cpu_samples[fname]
                    / stats.elapsed_time
                    # 100 * stats.cpu_samples[fname] / stats.total_cpu_samples
                )
            except ZeroDivisionError:
                percent_cpu_time = 0

            # Ignore files responsible for less than some percent of execution time and fewer than a threshold # of mallocs.
            if (
                sum(stats.memory_stats.memory_malloc_samples[fname].values())
                < self.malloc_threshold
                and percent_cpu_time < self.cpu_percent_threshold
            ):
                continue

            report_files.append(fname)

        # Don't actually output the profile if we are a child process.
        # Instead, write info to disk for the main process to collect.
        if pid:
            stats.output_stats(pid, python_alias_dir)
            # Return a value to indicate that the stats were successfully
            # output to the proper directory
            return {"is_child": True}

        if len(report_files) == 0:
            return {}

        for fname in report_files:

            # If the file was actually a Jupyter (IPython) cell,
            # restore its name, as in "[12]".
            fname_print = fname

            result = re.match(r"_ipython-input-([0-9]+)-.*", fname_print)
            if result:
                fname_print = Filename("[" + result.group(1) + "]")

            # Leak analysis
            # First, compute AVERAGE memory consumption.
            avg_mallocs: Dict[LineNumber, float] = defaultdict(float)
            for line_no in stats.memory_stats.memory_malloc_count[fname]:
                n_malloc_mb = stats.memory_stats.memory_aggregate_footprint[fname][
                    line_no
                ]
                count = stats.memory_stats.memory_malloc_count[fname][line_no]
                if count:
                    avg_mallocs[line_no] = n_malloc_mb / count
                else:
                    # Setting to n_malloc_mb addresses the edge case where this allocation is the last line executed.
                    avg_mallocs[line_no] = n_malloc_mb

            avg_mallocs = dict(
                sorted(avg_mallocs.items(), key=itemgetter(1), reverse=True)
            )

            # Now only report potential leaks if the allocation
            # velocity (growth rate) is above some threshold.
            leaks = ScaleneLeakAnalysis.compute_leaks(
                growth_rate, stats, avg_mallocs, fname
            )

            # Sort in descending order by least likelihood
            leaks = sorted(leaks, key=itemgetter(1), reverse=True)

            reported_leaks = {}

            for leak_lineno, leak_likelihood, leak_velocity in leaks:
                reported_leaks[str(leak_lineno)] = {
                    "likelihood": leak_likelihood,
                    "velocity_mb_s": leak_velocity / stats.elapsed_time,
                }

            # Print header.
            if not stats.cpu_stats.total_cpu_samples:
                percent_cpu_time = 0
            else:
                percent_cpu_time = (
                    100
                    * stats.cpu_stats.cpu_samples[fname]
                    / stats.cpu_stats.total_cpu_samples
                )

            # Print out the the profile for the source, line by line.
            full_fname = fname
            try:
                with open(full_fname, "r", encoding="utf-8") as source_file:
                    code_lines = source_file.readlines()

            except (FileNotFoundError, OSError):
                continue
            # Find all enclosing regions (loops or function defs) for each line of code.

            code_str = "".join(code_lines)

            enclosing_regions = ScaleneAnalysis.find_regions(code_str)
            outer_loop = ScaleneAnalysis.find_outermost_loop(code_str)
            imports = ScaleneAnalysis.get_native_imported_modules(code_str)

            output["files"][fname_print] = {
                "percent_cpu_time": percent_cpu_time,
                "lines": [],
                "leaks": reported_leaks,
                "imports": imports,
            }
            for lineno, line in enumerate(code_lines, start=1):
                # Protect against JS 'injection' in Python comments by replacing some characters with Unicode.
                # This gets unescaped in scalene-gui.js.
                line = line.replace("&", "\\u0026")
                line = line.replace("<", "\\u003c")
                line = line.replace(">", "\\u003e")
                profile_line = self.output_profile_line(
                    fname=fname,
                    fname_print=fname_print,
                    line_no=LineNumber(lineno),
                    line=line,
                    stats=stats,
                    profile_this_code=profile_this_code,
                    profile_memory=profile_memory,
                    force_print=False,
                )
                if profile_line:
                    profile_line["start_region_line"] = enclosing_regions[lineno][0]
                    profile_line["end_region_line"] = enclosing_regions[lineno][1]
                    profile_line["start_outermost_loop"] = outer_loop[lineno][0]
                    profile_line["end_outermost_loop"] = outer_loop[lineno][1]

                    try:
                        LineDetail(**profile_line)
                    except ValidationError as e:
                        print("Warning: JSON failed validation:")
                        print(e)

                    # When reduced-profile set, only output if the payload for the line is non-zero.
                    if reduced_profile:
                        profile_line_copy = copy.copy(profile_line)
                        del profile_line_copy["line"]
                        del profile_line_copy["lineno"]
                        if not any(profile_line_copy.values()):
                            continue
                    output["files"][fname_print]["lines"].append(profile_line)

            fn_stats = stats.build_function_stats(fname)
            # Check CPU samples and memory samples.
            print_fn_summary = False
            all_samples = set()
            all_samples |= set(fn_stats.cpu_stats.cpu_samples_python.keys())
            all_samples |= set(fn_stats.cpu_stats.cpu_samples_c.keys())
            all_samples |= set(fn_stats.memory_stats.memory_malloc_samples.keys())
            all_samples |= set(fn_stats.memory_stats.memory_free_samples.keys())
            all_samples |= set(fn_stats.gpu_stats.gpu_samples.keys())
            print_fn_summary = any(fn != fname for fn in all_samples)
            output["files"][fname_print]["functions"] = []
            if print_fn_summary:
                for fn_name in sorted(
                    all_samples,
                    key=lambda k: stats.firstline_map[k],
                ):
                    if fn_name == fname:
                        continue
                    profile_line = self.output_profile_line(
                        fname=fn_name,
                        fname_print=fn_name,
                        # line 1 is where function stats are
                        # accumulated; see
                        # ScaleneStatistics.build_function_stats
                        line_no=LineNumber(1),
                        line=fn_name,  # Set the source line to just the function name.
                        stats=fn_stats,
                        profile_this_code=profile_this_code,
                        profile_memory=profile_memory,
                        force_print=True,
                    )
                    if profile_line:
                        # Fix the line number to point to the first line of the function.
                        profile_line["lineno"] = stats.firstline_map[fn_name]
                        output["files"][fname_print]["functions"].append(profile_line)

        # Validate the schema
        try:
            ScaleneJSONSchema(**output)
        except ValidationError as e:
            print("Warning: JSON failed validation:")
            print(e)
        return output



#!/usr/bin/env python3

import builtins

try:
    builtins.profile
except AttributeError:
    # No line profiler, provide a pass-through version
    def profile(func):
        return func

    builtins.profile = profile

import numpy as np

# import math

# from numpy import linalg as LA

arr = [i for i in range(1, 1000)]


@profile
def doit1(x):
    #    x = [i*i for i in range(1,1000)][0]
    y = 1
    #    w, v = LA.eig(np.diag(arr)) # (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)))
    x = [i * i for i in range(0, 100000)][99999]
    y1 = [i * i for i in range(0, 200000)][199999]
    z1 = [i for i in range(0, 300000)][299999]
    z = x * y
    #    z = np.multiply(x, y)
    return z


def doit2(x):
    i = 0
    #    zarr = [math.cos(13) for i in range(1,100000)]
    #    z = zarr[0]
    z = 0.1
    while i < 100000:
        #        z = math.cos(13)
        #        z = np.multiply(x,x)
        #        z = np.multiply(z,z)
        #        z = np.multiply(z,z)
        z = z * z
        z = x * x
        z = z * z
        z = z * z
        i += 1
    return z


@profile
def doit3(x):
    for i in range(1000000):
        z = x + 1
        z = x + 1
        z = x + 1
        z = x + z
        z = x + z
    #    z = np.cos(x)
    return z


def stuff():
    #    y = np.random.randint(1, 100, size=50000000)[49999999]
    x = 1.01
    for i in range(1, 3):
        # print(i)
        for j in range(1, 3):
            x = doit1(x)
            x = doit2(x)
            x = doit3(x)
            x = 1.01
    return x


import sys

# print("TESTME")
# print(sys.argv)
stuff()



from setuptools import setup, find_packages
from setuptools.extension import Extension
from os import path, environ
import sys
import sysconfig
from pathlib import Path

# needed for isolated environment
sys.path.insert(0, str(Path(__file__).parent.resolve()))
from scalene.scalene_config import scalene_version

sys.path.pop(0)


if sys.platform == "darwin":
    import sysconfig

    mdt = "MACOSX_DEPLOYMENT_TARGET"
    target = environ[mdt] if mdt in environ else sysconfig.get_config_var(mdt)
    # target >= 10.9 is required for gcc/clang to find libstdc++ headers
    if [int(n) for n in target.split(".")] < [10, 9]:
        from os import execve

        newenv = environ.copy()
        newenv[mdt] = "10.9"
        execve(sys.executable, [sys.executable] + sys.argv, newenv)


def compiler_archs(compiler: str):
    """Discovers what platforms the given compiler supports; intended for MacOS use"""
    import tempfile
    import subprocess

    print(f"Compiler: {compiler}")
    arch_flags = []

    # see also the architectures tested for in .github/workflows/build-and-upload.yml
    for arch in ["x86_64", "arm64", "arm64e"]:
        with tempfile.TemporaryDirectory() as tmpdir:
            cpp = Path(tmpdir) / "test.cxx"
            cpp.write_text("int main() {return 0;}\n")
            out = Path(tmpdir) / "a.out"
            p = subprocess.run(
                [compiler, "-arch", arch, str(cpp), "-o", str(out)], capture_output=True
            )
            if p.returncode == 0:
                arch_flags += ["-arch", arch]

    print(f"Discovered {compiler} arch flags: {arch_flags}")
    return arch_flags


def extra_compile_args():
    """Returns extra compiler args for platform."""
    if sys.platform == "win32":
        return ["/std:c++14"]  # for Visual Studio C++

    return ["-std=c++14"]


def make_command():
    #    return 'nmake' if sys.platform == 'win32' else 'make'  # 'nmake' isn't found on github actions' VM
    return "make"


def dll_suffix():
    """Returns the file suffix ("extension") of a DLL"""
    if sys.platform == "win32":
        return ".dll"
    if sys.platform == "darwin":
        return ".dylib"
    return ".so"


def read_file(name):
    """Returns a file's contents"""
    with open(path.join(path.dirname(__file__), name), encoding="utf-8") as f:
        return f.read()


import setuptools.command.egg_info


class EggInfoCommand(setuptools.command.egg_info.egg_info):
    """Custom command to download vendor libs before creating the egg_info."""

    def run(self):
        if sys.platform != "win32":
            self.spawn([make_command(), "vendor-deps"])
        super().run()


# Force building platform-specific wheel to avoid the Windows wheel
# (which doesn't include libscalene, and thus would be considered "pure")
# being used for other platforms.
try:
    from wheel.bdist_wheel import bdist_wheel

    class BdistWheelCommand(bdist_wheel):
        def finalize_options(self):
            super().finalize_options()
            self.root_is_pure = False

except (ImportError, ModuleNotFoundError):
    # Disable wheel if `wheel` not installed.
    print(
        "If this installation does not work, run `pip install setuptools wheel` and try again."
    )
    BdistWheelCommand = None

import setuptools.command.build_ext


class BuildExtCommand(setuptools.command.build_ext.build_ext):
    """Custom command that runs 'make' to generate libscalene, and also does MacOS
    supported --arch flag discovery."""

    def build_extensions(self):
        arch_flags = []
        if sys.platform == "darwin":
            # The only sure way to tell which compiler build_ext is going to use
            # seems to be to customize a build_ext and look at its internal flags :(
            # Also, note that self.plat_name here isn't "...-universal2" even if that
            # is what we're building; that's only in bdist_wheel.plat_name.
            arch_flags += compiler_archs(self.compiler.compiler_cxx[0])
            for ext in self.extensions:
                # While the flags _could_ be different between the programs used for
                # C and C++ compilation and linking, we have no way to adapt them here,
                # so it seems best to just use them and let it error out if not recognized.
                ext.extra_compile_args += arch_flags
                ext.extra_link_args += arch_flags

        super().build_extensions()

        # No build of DLL for Windows currently.
        if sys.platform != "win32":
            self.build_libscalene(arch_flags)  # XXX should we pass compiler_cxx here?

    def build_libscalene(self, arch_flags):
        scalene_temp = path.join(self.build_temp, "scalene")
        scalene_lib = path.join(self.build_lib, "scalene")
        libscalene = "libscalene" + dll_suffix()
        self.mkpath(scalene_temp)
        self.mkpath(scalene_lib)
        self.spawn(
            [make_command(), "OUTDIR=" + scalene_temp, "ARCH=" + " ".join(arch_flags)]
        )
        self.copy_file(
            path.join(scalene_temp, libscalene), path.join(scalene_lib, libscalene)
        )

    def copy_extensions_to_source(self):
        # self.inplace is temporarily overridden while running build_extensions,
        # so inplace copying (for pip install -e, setup.py develop) must be done here.

        super().copy_extensions_to_source()

        if sys.platform != "win32":
            scalene_lib = path.join(self.build_lib, "scalene")
            inplace_dir = self.get_finalized_command("build_py").get_package_dir(
                "scalene"
            )
            libscalene = "libscalene" + dll_suffix()
            self.copy_file(
                path.join(scalene_lib, libscalene), path.join(inplace_dir, libscalene)
            )


get_line_atomic = Extension(
    "scalene.get_line_atomic",
    include_dirs=[".", "vendor/Heap-Layers", "vendor/Heap-Layers/utility"],
    sources=["src/source/get_line_atomic.cpp"],
    extra_compile_args=extra_compile_args(),
    py_limited_api=True,  # for binary compatibility
    language="c++",
)

pywhere = Extension(
    "scalene.pywhere",
    include_dirs=[".", "src", "src/include"],
    depends=["src/include/traceconfig.hpp"],
    sources=["src/source/pywhere.cpp", "src/source/traceconfig.cpp"],
    extra_compile_args=extra_compile_args(),
    py_limited_api=False,
    language="c++",
)

# If we're testing packaging, build using a ".devN" suffix in the version number,
# so that we can upload new files (as testpypi/pypi don't allow re-uploading files with
# the same name as previously uploaded).
# Numbering scheme: https://www.python.org/dev/peps/pep-0440
dev_build = (".dev" + environ["DEV_BUILD"]) if "DEV_BUILD" in environ else ""


def bdist_wheel_options():
    if sys.platform == "darwin":
        # Build universal wheels on MacOS.
        # ---
        # On MacOS >= 11, all builds are compatible within a major MacOS version, so Python "floors"
        # all minor versions to 0, leading to tags like like "macosx_11_0_universal2". If you use
        # the actual (non-0) minor name in the build platform, it isn't recognized.
        # ---
        # It would be nice to check whether we're actually building multi-architecture,
        # but that depends on the platforms supported by the compiler build_ext wants to use,
        # which is hard to obtain (see BuildExtCommand above).
        import platform

        v = platform.mac_ver()[0]
        major = int(v.split(".")[0])
        if major >= 11:
            v = f"{major}.0"
        return {"plat_name": f"macosx-{v}-universal2"}

    return {}


setup(
    version=scalene_version + dev_build,
    packages=find_packages(),
    cmdclass={
        "bdist_wheel": BdistWheelCommand,
        "egg_info": EggInfoCommand,
        "build_ext": BuildExtCommand,
    },
    ext_modules=([get_line_atomic, pywhere] if sys.platform != "win32" else []),
    include_package_data=True,
    options={"bdist_wheel": bdist_wheel_options()},
)



# file scalene/launchbrowser.py:25-46
# lines [27, 30, 31]
# branches ['26->27', '28->30', '30->31', '30->34']

import pytest
import platform
from unittest.mock import patch
from scalene.launchbrowser import launch_browser_insecure


@pytest.fixture
def mock_platform_system():
    with patch("platform.system") as mock:
        yield mock


def test_launch_browser_insecure_on_mac(mock_platform_system):
    mock_platform_system.return_value = "Darwin"
    with patch("webbrowser.register") as mock_register, patch(
        "webbrowser.get"
    ) as mock_get, patch("tempfile.TemporaryDirectory") as mock_temp_dir:
        mock_temp_dir.return_value.__enter__.return_value = "/tmp"
        launch_browser_insecure("http://example.com")
        mock_register.assert_called_once()
        mock_get.assert_called_once()
        assert mock_get.call_args[0][0].startswith(
            "/Applications/Google\\ Chrome.app/Contents/MacOS/Google\\ Chrome"
        )


def test_launch_browser_insecure_on_windows(mock_platform_system):
    mock_platform_system.return_value = "Windows"
    with patch("webbrowser.register") as mock_register, patch(
        "webbrowser.get"
    ) as mock_get, patch("tempfile.TemporaryDirectory") as mock_temp_dir:
        mock_temp_dir.return_value.__enter__.return_value = "C:\\Temp"
        launch_browser_insecure("http://example.com")
        mock_register.assert_called_once()
        mock_get.assert_called_once()
        assert mock_get.call_args[0][0].startswith(
            "C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe"
        )



#!/usr/bin/env python3
import numpy as np

# import math

# from numpy import linalg as LA

arr = [i for i in range(1, 1000)]


def doit1(x):
    y = 1
    x = [i * i for i in range(0, 100000)][99999]
    y1 = [i * i for i in range(0, 200000)][199999]
    z1 = [i for i in range(0, 300000)][299999]
    z = x * y * y1 * z1
    return z


def doit2(x):
    i = 0
    z = 0.1
    while i < 100000:
        z = z * z
        z = x * x
        z = z * z
        z = z * z
        i += 1
    return z


def doit3(x):
    z = x + 1
    z = x + 1
    z = x + 1
    z = x + z
    z = x + z
    return z


def stuff():
    #    y = np.random.randint(1, 100, size=50000000)[49999999]
    x = 1.01
    for i in range(1, 10):
        print(i)
        for j in range(1, 10):
            x = doit1(x)
            x = doit2(x)
            x = doit3(x)
            x = 1.01
    return x


import sys

print("TESTME")
print(sys.argv)
stuff()



import sys
import subprocess
import tempfile

dir = tempfile.TemporaryDirectory()
cmd = [
    sys.executable,
    "-m",
    "scalene",
    "--cli",
    "--outfile",
    dir.name,
    "../testme.py",
    "--cpu-only",
]

print(cmd)
print(
    f"If bug 691 is fixed, you will see \n    scalene: error: outfile {dir.name} is a directory"
)
proc = subprocess.run(cmd)



#!/usr/bin/env python3
import time
import pyperf

from sympy import expand, symbols, integrate, tan, summation
from sympy.core.cache import clear_cache


def bench_expand():
    x, y, z = symbols("x y z")
    expand((1 + x + y + z) ** 20)


def bench_integrate():
    x, y = symbols("x y")
    f = (1 / tan(x)) ** 10
    return integrate(f, x)


def bench_sum():
    x, i = symbols("x i")
    summation(x**i / i, (i, 1, 400))


def bench_str():
    x, y, z = symbols("x y z")
    str(expand((x + 2 * y + 3 * z) ** 30))


def bench_sympy(loops, func):
    timer = pyperf.perf_counter
    dt = 0

    for _ in range(loops):
        # Don't benchmark clear_cache(), exclude it of the benchmark
        clear_cache()

        t0 = timer()
        func()
        dt += timer() - t0

    return dt


BENCHMARKS = ("expand", "integrate", "sum", "str")


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


if __name__ == "__main__":
    # sympy-expand

    import gc

    gc.disable()

    start_p = time.perf_counter()
    for _ in range(25):
        # Don't benchmark clear_cache(), exclude it of the benchmark
        clear_cache()
        bench_expand()

    stop_p = time.perf_counter()
    print("Time elapsed: ", stop_p - start_p)



from __future__ import annotations

import os
import pathlib
import pickle
import time
from collections import defaultdict
from functools import total_ordering
from pydantic import PositiveInt
from typing import (
    Any,
    Dict,
    List,
    NewType,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import cloudpickle

from scalene.runningstats import RunningStats

Address = NewType("Address", str)
Filename = NewType("Filename", str)
LineNumber = NewType("LineNumber", PositiveInt)
ByteCodeIndex = NewType("ByteCodeIndex", int)
T = TypeVar("T")


class ProfilingSample:
    def __init__(
        self,
        action: str,
        alloc_time: int,
        count: float,
        python_fraction: float,
        pointer: "Address",
        filename: "Filename",
        lineno: "LineNumber",
        bytecode_index: "ByteCodeIndex",
    ) -> None:
        self.action = action
        self.alloc_time = alloc_time
        self.count = count
        self.python_fraction = python_fraction
        self.pointer = pointer
        self.filename = filename
        self.lineno = lineno
        self.bytecode_index = bytecode_index


@total_ordering
class MemcpyProfilingSample:
    def __init__(
        self,
        memcpy_time: int,
        count: float,
        filename: "Filename",
        lineno: "LineNumber",
        bytecode_index: "ByteCodeIndex",
    ) -> None:
        self.memcpy_time = memcpy_time
        self.count = count
        self.filename = filename
        self.lineno = lineno
        self.bytecode_index = bytecode_index

    def __lt__(self, other: "MemcpyProfilingSample") -> bool:
        """Compare based on memcpy_time for sorting."""
        return self.memcpy_time < other.memcpy_time

    def __eq__(self, other: object) -> bool:
        """Compare equality based on all fields."""
        if not isinstance(other, MemcpyProfilingSample):
            return NotImplemented
        return (
            self.memcpy_time == other.memcpy_time
            and self.count == other.count
            and self.filename == other.filename
            and self.lineno == other.lineno
            and self.bytecode_index == other.bytecode_index
        )


class StackFrame:
    """Represents a single frame in the stack."""

    def __init__(self, filename: str, function_name: str, line_number: int) -> None:
        self.filename = filename
        self.function_name = function_name
        self.line_number = line_number

    def __repr__(self) -> str:
        return f"{self.filename} {self.function_name}:{self.line_number};"

    def __hash__(self) -> int:
        return hash((self.filename, self.function_name, self.line_number))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, StackFrame):
            return False
        return (
            self.filename == other.filename
            and self.function_name == other.function_name
            and self.line_number == other.line_number
        )


class StackStats:
    """Represents statistics for a stack."""

    def __init__(
        self, count: int, python_time: float, c_time: float, cpu_samples: float
    ) -> None:
        self.count = count
        self.python_time = python_time
        self.c_time = c_time
        self.cpu_samples = cpu_samples

    def __repr__(self) -> str:
        return f" {self.count}"


class CPUStatistics:
    """Statistics related to CPU usage."""

    def __init__(self) -> None:
        # CPU samples for each location in the program spent in the interpreter
        self.cpu_samples_python: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # CPU samples for each location in the program spent in C / libraries / system calls
        self.cpu_samples_c: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # Running stats for the fraction of time running on the CPU
        self.cpu_utilization: Dict[Any, Dict[Any, RunningStats]] = defaultdict(
            lambda: defaultdict(RunningStats)
        )

        # Running stats for core utilization
        self.core_utilization: Dict[Any, Dict[Any, RunningStats]] = defaultdict(
            lambda: defaultdict(RunningStats)
        )

        # Running count of total CPU samples per file. Used to prune reporting
        self.cpu_samples: Dict[Any, float] = defaultdict(float)

        # How many CPU samples have been collected
        self.total_cpu_samples: float = 0.0

        self.cpu_samples_list: Dict[Any, Dict[Any, List[float]]] = defaultdict(
            lambda: defaultdict(list)
        )

    def clear(self) -> None:
        """Reset all CPU statistics."""
        self.cpu_samples_python.clear()
        self.cpu_samples_c.clear()
        self.cpu_samples_list.clear()
        self.cpu_utilization.clear()
        self.core_utilization.clear()
        self.cpu_samples.clear()
        self.total_cpu_samples = 0.0


class MemoryStatistics:
    """Statistics related to memory usage."""

    def __init__(self) -> None:
        # Total allocation samples taken
        self.alloc_samples: int = 0

        # Running count of malloc samples per file. Used to prune reporting
        self.malloc_samples: Dict[Any, float] = defaultdict(float)

        # malloc samples for each location in the program
        self.memory_malloc_samples: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # number of times samples were added for the above
        self.memory_malloc_count: Dict[Any, Dict[Any, int]] = defaultdict(
            lambda: defaultdict(int)
        )

        # the current footprint for this line
        self.memory_current_footprint: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # the max footprint for this line
        self.memory_max_footprint: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # the current high watermark for this line
        self.memory_current_highwater_mark: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # the aggregate footprint for this line (sum of all final "current"s)
        self.memory_aggregate_footprint: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # the last malloc to trigger a sample (used for leak detection)
        self.last_malloc_triggered = (Filename(""), LineNumber(0), Address("0x0"))

        # mallocs attributable to Python, for each location in the program
        self.memory_python_samples: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # free samples for each location in the program
        self.memory_free_samples: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # number of times samples were added for the above
        self.memory_free_count: Dict[Any, Dict[Any, int]] = defaultdict(
            lambda: defaultdict(int)
        )

        # memcpy samples for each location in the program
        self.memcpy_samples: Dict[Any, Dict[Any, int]] = defaultdict(
            lambda: defaultdict(int)
        )

        # leak score tracking
        self.leak_score: Dict[Any, Dict[Any, Tuple[int, int]]] = defaultdict(
            lambda: defaultdict(lambda: ((0, 0)))
        )

        self.allocation_velocity: Tuple[float, float] = (0.0, 0.0)

        # Total memory samples
        self.total_memory_malloc_samples: float = 0.0
        self.total_memory_free_samples: float = 0.0

        # Memory footprint
        self.current_footprint: float = 0.0
        self.max_footprint: float = 0.0
        self.max_footprint_python_fraction: float = 0
        self.max_footprint_loc: Optional[Tuple["Filename", "LineNumber"]] = None

        # Memory footprint samples (time, footprint)
        self.memory_footprint_samples: List[Tuple[float, float]] = []

        # Same, but per line
        self.per_line_footprint_samples: Dict[Any, Dict[Any, List[Any]]] = defaultdict(
            lambda: defaultdict(list)
        )

    def clear(self) -> None:
        """Reset all memory statistics except for memory footprint."""
        self.alloc_samples = 0
        self.malloc_samples.clear()
        self.memory_malloc_samples.clear()
        self.memory_malloc_count.clear()
        self.memory_current_footprint.clear()
        self.memory_max_footprint.clear()
        self.memory_current_highwater_mark.clear()
        self.memory_aggregate_footprint.clear()
        self.memory_python_samples.clear()
        self.memory_free_samples.clear()
        self.memory_free_count.clear()
        self.memcpy_samples.clear()
        self.total_memory_malloc_samples = 0.0
        self.total_memory_free_samples = 0.0
        self.current_footprint = 0.0
        self.leak_score.clear()
        self.last_malloc_triggered = (Filename(""), LineNumber(0), Address("0x0"))
        self.allocation_velocity = (0.0, 0.0)
        self.per_line_footprint_samples.clear()
        # Not clearing current footprint
        # Not clearing max footprint

    def clear_all(self) -> None:
        """Clear all memory statistics."""
        self.clear()
        self.current_footprint = 0
        self.max_footprint = 0
        self.max_footprint_loc = None
        self.per_line_footprint_samples.clear()


class GPUStatistics:
    """Statistics related to GPU usage."""

    def __init__(self) -> None:
        # GPU samples for each location in the program
        self.gpu_samples: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # Number of GPU samples taken (actually weighted by elapsed wallclock time)
        self.n_gpu_samples: Dict[Any, Dict[Any, float]] = defaultdict(
            lambda: defaultdict(float)
        )

        # GPU memory samples for each location in the program
        self.gpu_mem_samples: Dict[Any, Dict[Any, RunningStats]] = defaultdict(
            lambda: defaultdict(RunningStats)
        )

        # How many GPU samples have been collected
        self.total_gpu_samples: float = 0.0

    def clear(self) -> None:
        """Reset all GPU statistics."""
        self.gpu_samples.clear()
        self.n_gpu_samples.clear()
        self.gpu_mem_samples.clear()
        self.total_gpu_samples = 0.0


class ScaleneStatistics:
    @classmethod
    def _get_payload_contents(cls) -> List[str]:
        """Generate the list of payload contents programmatically."""
        contents = []

        # Memory statistics
        memory_stats = MemoryStatistics()
        for attr in dir(memory_stats):
            if not attr.startswith("_") and not callable(getattr(memory_stats, attr)):
                contents.append(f"memory_stats.{attr}")

        # CPU statistics
        cpu_stats = CPUStatistics()
        for attr in dir(cpu_stats):
            if not attr.startswith("_") and not callable(getattr(cpu_stats, attr)):
                contents.append(f"cpu_stats.{attr}")

        # GPU statistics
        gpu_stats = GPUStatistics()
        for attr in dir(gpu_stats):
            if not attr.startswith("_") and not callable(getattr(gpu_stats, attr)):
                contents.append(f"gpu_stats.{attr}")

        # General statistics (attributes directly on ScaleneStatistics)
        # Create a temporary instance to inspect attributes
        temp_stats = cls()
        for attr in dir(temp_stats):
            # Skip private attributes, callables, and attributes that are already covered by stats classes
            if (
                not attr.startswith("_")
                and not callable(getattr(temp_stats, attr))
                and not attr.startswith(("memory_stats", "cpu_stats", "gpu_stats"))
                and attr not in ("start_time", "payload_contents")
            ):
                contents.append(attr)

        # Ensure we have all the required fields in the correct order
        required_fields = [
            "memory_stats.max_footprint",
            "memory_stats.max_footprint_loc",
            "memory_stats.current_footprint",
            "elapsed_time",
            "memory_stats.alloc_samples",
            "stacks",
            "cpu_stats.total_cpu_samples",
            "cpu_stats.cpu_samples_c",
            "cpu_stats.cpu_samples_python",
            "bytei_map",
            "cpu_stats.cpu_samples",
            "cpu_stats.cpu_utilization",
            "cpu_stats.core_utilization",
            "memory_stats.memory_malloc_samples",
            "memory_stats.memory_python_samples",
            "memory_stats.memory_free_samples",
            "memory_stats.memcpy_samples",
            "memory_stats.memory_max_footprint",
            "memory_stats.per_line_footprint_samples",
            "memory_stats.total_memory_free_samples",
            "memory_stats.total_memory_malloc_samples",
            "memory_stats.memory_footprint_samples",
            "function_map",
            "firstline_map",
            "gpu_stats.gpu_samples",
            "gpu_stats.n_gpu_samples",
            "gpu_stats.gpu_mem_samples",
            "gpu_stats.total_gpu_samples",
            "memory_stats.memory_malloc_count",
            "memory_stats.memory_free_count",
        ]

        # Add any missing required fields
        for field in required_fields:
            if field not in contents:
                contents.append(field)

        # Sort to ensure consistent order
        contents.sort()

        return contents

    # Initialize payload_contents after class definition
    payload_contents: List[str] = []

    def __init__(self) -> None:
        # time the profiling started
        self.start_time: float = 0

        # total time spent in program being profiled
        self.elapsed_time: float = 0

        # absolute start time (time.time())
        self.start_time_absolute: float = 0

        # perf_counter start time
        self.start_time_perf: float = 0

        # full stacks taken during CPU samples, together with number of hits
        self.stacks: Dict[Any, StackStats] = defaultdict(
            lambda: StackStats(0, 0.0, 0.0, 0.0)
        )

        # Initialize statistics classes
        self.cpu_stats = CPUStatistics()
        self.memory_stats = MemoryStatistics()
        self.gpu_stats = GPUStatistics()

        # maps byte indices to line numbers (collected at runtime)
        # [filename][lineno] -> set(byteindex)
        self.bytei_map: Dict[Any, Dict[Any, Set["ByteCodeIndex"]]] = defaultdict(
            lambda: defaultdict(set)
        )

        # maps filenames and line numbers to functions (collected at runtime)
        # [filename][lineno] -> function name
        self.function_map: Dict[Any, Dict[Any, "Filename"]] = defaultdict(
            lambda: defaultdict(lambda: Filename(""))
        )
        self.firstline_map: Dict[Any, "LineNumber"] = defaultdict(lambda: LineNumber(1))

    def clear(self) -> None:
        """Reset all statistics except for memory footprint."""
        self.start_time = 0
        self.elapsed_time = 0
        self.stacks.clear()
        self.cpu_stats.clear()
        self.memory_stats.clear()
        self.gpu_stats.clear()
        self.bytei_map.clear()

    def clear_all(self) -> None:
        """Clear all statistics."""
        self.clear()
        self.memory_stats.clear_all()

    def start_clock(self) -> None:
        """Start the timer."""
        self.start_time = time.time()
        self.start_time_absolute = time.time()
        self.start_time_perf = time.perf_counter()

    def stop_clock(self) -> None:
        """Stop the timer."""
        if self.start_time > 0:
            self.elapsed_time += time.time() - self.start_time
        self.start_time = 0

    def build_function_stats(self, filename: "Filename") -> "ScaleneStatistics":
        """Produce aggregated statistics for each function."""
        fn_stats = ScaleneStatistics()
        fn_stats.elapsed_time = self.elapsed_time
        fn_stats.cpu_stats.total_cpu_samples = self.cpu_stats.total_cpu_samples
        fn_stats.gpu_stats.total_gpu_samples = self.gpu_stats.total_gpu_samples
        fn_stats.gpu_stats.n_gpu_samples = self.gpu_stats.n_gpu_samples
        fn_stats.memory_stats.total_memory_malloc_samples = (
            self.memory_stats.total_memory_malloc_samples
        )
        first_line_no = LineNumber(1)
        fn_stats.function_map = self.function_map
        fn_stats.firstline_map = self.firstline_map
        for line_no in self.function_map[filename]:
            fn_name = self.function_map[filename][line_no]
            if fn_name == "<module>":
                continue

            fn_stats.cpu_stats.cpu_samples_c[fn_name][
                first_line_no
            ] += self.cpu_stats.cpu_samples_c[filename][line_no]
            fn_stats.cpu_stats.cpu_samples_python[fn_name][
                first_line_no
            ] += self.cpu_stats.cpu_samples_python[filename][line_no]
            fn_stats.gpu_stats.gpu_samples[fn_name][
                first_line_no
            ] += self.gpu_stats.gpu_samples[filename][line_no]
            fn_stats.gpu_stats.n_gpu_samples[fn_name][
                first_line_no
            ] += self.gpu_stats.n_gpu_samples[filename][line_no]
            fn_stats.gpu_stats.gpu_mem_samples[fn_name][
                first_line_no
            ] += self.gpu_stats.gpu_mem_samples[filename][line_no]
            fn_stats.cpu_stats.cpu_utilization[fn_name][
                first_line_no
            ] += self.cpu_stats.cpu_utilization[filename][line_no]
            fn_stats.cpu_stats.core_utilization[fn_name][
                first_line_no
            ] += self.cpu_stats.core_utilization[filename][line_no]
            fn_stats.memory_stats.per_line_footprint_samples[fn_name][
                first_line_no
            ] += self.memory_stats.per_line_footprint_samples[filename][line_no]
            fn_stats.memory_stats.memory_malloc_count[fn_name][
                first_line_no
            ] += self.memory_stats.memory_malloc_count[filename][line_no]
            fn_stats.memory_stats.memory_free_count[fn_name][
                first_line_no
            ] += self.memory_stats.memory_free_count[filename][line_no]
            fn_stats.memory_stats.memory_malloc_samples[fn_name][
                first_line_no
            ] += self.memory_stats.memory_malloc_samples[filename][line_no]
            fn_stats.memory_stats.memory_python_samples[fn_name][
                first_line_no
            ] += self.memory_stats.memory_python_samples[filename][line_no]
            fn_stats.memory_stats.memory_free_samples[fn_name][
                first_line_no
            ] += self.memory_stats.memory_free_samples[filename][line_no]
            for index in self.bytei_map[filename][line_no]:
                fn_stats.bytei_map[fn_name][first_line_no].add(
                    ByteCodeIndex(index)  # was 0
                )
            fn_stats.memory_stats.memcpy_samples[fn_name][
                first_line_no
            ] += self.memory_stats.memcpy_samples[filename][line_no]
            fn_stats.memory_stats.leak_score[fn_name][first_line_no] = (
                fn_stats.memory_stats.leak_score[fn_name][first_line_no][0]
                + self.memory_stats.leak_score[filename][line_no][0],
                fn_stats.memory_stats.leak_score[fn_name][first_line_no][1]
                + self.memory_stats.leak_score[filename][line_no][1],
            )
            fn_stats.memory_stats.memory_max_footprint[fn_name][first_line_no] = max(
                fn_stats.memory_stats.memory_max_footprint[fn_name][first_line_no],
                self.memory_stats.memory_max_footprint[filename][line_no],
            )
            fn_stats.memory_stats.memory_aggregate_footprint[fn_name][
                first_line_no
            ] += self.memory_stats.memory_aggregate_footprint[filename][line_no]

        return fn_stats

    def output_stats(self, pid: int, dir_name: pathlib.Path) -> None:
        """Output statistics for a particular process to a given directory."""
        payload: List[Any] = []
        for attr_path in ScaleneStatistics.payload_contents:
            # Split the path into parts
            parts = attr_path.split(".")
            # Start with self
            value = self
            # Traverse the path
            for part in parts:
                value = getattr(value, part)
            payload.append(value)

        # Create a file in the Python alias directory with the relevant info.
        out_filename = os.path.join(dir_name, f"scalene{pid}-{str(os.getpid())}")
        with open(out_filename, "wb") as out_file:
            cloudpickle.dump(payload, out_file)

    @staticmethod
    def increment_per_line_samples(
        dest: Dict["Filename", Dict["LineNumber", T]],
        src: Dict["Filename", Dict["LineNumber", T]],
    ) -> None:
        """Increment single-line dest samples by their value in src."""
        for filename in src:
            for lineno in src[filename]:
                v = src[filename][lineno]
                dest[filename][lineno] += v  # type: ignore

    @staticmethod
    def increment_cpu_utilization(
        dest: Dict["Filename", Dict["LineNumber", RunningStats]],
        src: Dict["Filename", Dict["LineNumber", RunningStats]],
    ) -> None:
        """Increment CPU utilization."""
        for filename in src:
            for lineno in src[filename]:
                dest[filename][lineno] += src[filename][lineno]

    @staticmethod
    def increment_core_utilization(
        dest: Dict["Filename", Dict["LineNumber", RunningStats]],
        src: Dict["Filename", Dict["LineNumber", RunningStats]],
    ) -> None:
        """Increment core utilization."""
        for filename in src:
            for lineno in src[filename]:
                dest[filename][lineno] += src[filename][lineno]

    def merge_stats(self, the_dir_name: pathlib.Path) -> None:
        """Merge all statistics in a given directory."""
        the_dir = pathlib.Path(the_dir_name)
        for f in list(the_dir.glob(os.path.join("**", "scalene*"))):
            # Skip empty files.
            if os.path.getsize(f) == 0:
                continue
            with open(f, "rb") as file:
                unpickler = pickle.Unpickler(file)
                try:
                    value = unpickler.load()
                except EOFError:
                    # Empty file for some reason.
                    continue
                x = ScaleneStatistics()
                for i, n in enumerate(ScaleneStatistics.payload_contents):
                    # Split the path into parts
                    parts = n.split(".")
                    # Start with x
                    target = x
                    # Traverse the path except the last part
                    for part in parts[:-1]:
                        target = getattr(target, part)
                    # Set the value
                    setattr(target, parts[-1], value[i])

                if x.memory_stats.max_footprint > self.memory_stats.max_footprint:
                    self.memory_stats.max_footprint = x.memory_stats.max_footprint
                    self.memory_stats.max_footprint_loc = (
                        x.memory_stats.max_footprint_loc
                    )
                self.memory_stats.current_footprint = max(
                    self.memory_stats.current_footprint,
                    x.memory_stats.current_footprint,
                )
                self.cpu_stats.cpu_utilization.update(x.cpu_stats.cpu_utilization)
                self.cpu_stats.core_utilization.update(x.cpu_stats.core_utilization)
                self.elapsed_time = max(self.elapsed_time, x.elapsed_time)
                self.memory_stats.alloc_samples += x.memory_stats.alloc_samples
                self.stacks.update(x.stacks)
                self.cpu_stats.total_cpu_samples += x.cpu_stats.total_cpu_samples
                self.gpu_stats.total_gpu_samples += x.gpu_stats.total_gpu_samples

                # Restore per-line sample increments
                self.increment_per_line_samples(
                    self.cpu_stats.cpu_samples_c, x.cpu_stats.cpu_samples_c
                )
                self.increment_per_line_samples(
                    self.cpu_stats.cpu_samples_python, x.cpu_stats.cpu_samples_python
                )
                self.increment_per_line_samples(
                    self.gpu_stats.gpu_samples, x.gpu_stats.gpu_samples
                )
                self.increment_per_line_samples(
                    self.gpu_stats.n_gpu_samples, x.gpu_stats.n_gpu_samples
                )
                self.increment_per_line_samples(
                    self.gpu_stats.gpu_mem_samples, x.gpu_stats.gpu_mem_samples
                )
                self.increment_per_line_samples(
                    self.memory_stats.memcpy_samples, x.memory_stats.memcpy_samples
                )
                self.increment_per_line_samples(
                    self.memory_stats.per_line_footprint_samples,
                    x.memory_stats.per_line_footprint_samples,
                )
                # Sorting each of the per_line_footprint_sample lists by time
                for filename in self.memory_stats.per_line_footprint_samples:
                    for lineno in self.memory_stats.per_line_footprint_samples[
                        filename
                    ]:
                        self.memory_stats.per_line_footprint_samples[filename][
                            lineno
                        ].sort(key=lambda x: x[0])
                self.increment_per_line_samples(
                    self.memory_stats.memory_malloc_count,
                    x.memory_stats.memory_malloc_count,
                )
                self.increment_per_line_samples(
                    self.memory_stats.memory_malloc_samples,
                    x.memory_stats.memory_malloc_samples,
                )
                self.increment_per_line_samples(
                    self.memory_stats.memory_python_samples,
                    x.memory_stats.memory_python_samples,
                )
                self.increment_per_line_samples(
                    self.memory_stats.memory_free_samples,
                    x.memory_stats.memory_free_samples,
                )
                self.increment_per_line_samples(
                    self.memory_stats.memory_free_count,
                    x.memory_stats.memory_free_count,
                )

                # Restore bytei_map and memory_max_footprint handling
                for filename in x.bytei_map:
                    for lineno in x.bytei_map[filename]:
                        v = x.bytei_map[filename][lineno]
                        self.bytei_map[filename][lineno] |= v
                        self.memory_stats.memory_max_footprint[filename][lineno] = max(
                            self.memory_stats.memory_max_footprint[filename][lineno],
                            x.memory_stats.memory_max_footprint[filename][lineno],
                        )

                # Restore cpu_samples handling
                for filename in x.cpu_stats.cpu_samples:
                    self.cpu_stats.cpu_samples[filename] += x.cpu_stats.cpu_samples[
                        filename
                    ]

                self.memory_stats.total_memory_free_samples += (
                    x.memory_stats.total_memory_free_samples
                )
                self.memory_stats.total_memory_malloc_samples += (
                    x.memory_stats.total_memory_malloc_samples
                )
                self.memory_stats.memory_footprint_samples += (
                    x.memory_stats.memory_footprint_samples
                )
                self.memory_stats.memory_footprint_samples.sort(key=lambda x: x[0])

                # Restore careful function_map handling
                for k, val in x.function_map.items():
                    if k in self.function_map:
                        self.function_map[k].update(val)
                    else:
                        self.function_map[k] = val
                self.firstline_map.update(x.firstline_map)
            os.remove(f)


# Initialize payload_contents after class definition
ScaleneStatistics.payload_contents = ScaleneStatistics._get_payload_contents()



# file scalene/scalene_parseargs.py:35-38
# lines [35, 36, 38]
# branches []

import pytest
from scalene.scalene_parseargs import ScaleneParseArgs


class StopJupyterExecution(Exception):
    pass


def test_clean_exit(monkeypatch):
    monkeypatch.setattr(
        "scalene.scalene_parseargs.StopJupyterExecution", StopJupyterExecution
    )
    with pytest.raises(StopJupyterExecution):
        ScaleneParseArgs.clean_exit()
